#include "IRT10em.h"

// New version distributed by JLV 3/16/99
// clean-up dt 5/20/99
// reduced to mins for CW 8/02 after expansion for IRTLRDIF by dt
// further reduced / made more parallel to IRT04s for 285 10/04
// mod 10/10 to remove LstArray class references

#define kZeroByte 0 // this is used in the code to set the null bit that ends C strings

int IRT10em::size = 1;		// default size of IRT10ems is 1
int IRT10em::nqpt = 46;		// default # quad pts is 46 / changed to 61 for irtlrdif 11/00
double IRT10em::qpmax = 4.5;	// also static; quadrature is set
								// to -4.5, -4.3, ... 4.5 // changed to 6 for irtlrdif 11/00
valarray<double> IRT10em::popdist;
valarray<double> IRT10em::theta;

//default constructor
IRT10em::IRT10em() {
	int i;

// set the statics:
	IRT10em::popdist.resize(IRT10em::nqpt);
	IRT10em::theta.resize(IRT10em::nqpt);

// the following six lines create the lists of IRT10em::theta (quadrature points)
// and the ordinates of the normal distribution at those points
	double incr = (2 * qpmax)/(nqpt-1);
	IRT10em::theta[0] = -qpmax;
	for (i = 1; i < nqpt; i++)
		IRT10em::theta[i] = IRT10em::theta[i-1] + incr;
	IRT10em::popdist = exp(-(((IRT10em::theta-0.0)*(IRT10em::theta-0.0))/(2*1.0*1.0)));
	IRT10em::popdist = IRT10em::popdist/IRT10em::popdist.sum(); 

	nitem = size;
	maxscore = size + 1;
	for (i = 0; i < 64; i++)
		DatFile[i] = kZeroByte;
	I = new ItemStuff [nitem];
	lk.resize(maxscore); 

	aCol = 1;
	bCol = 2;
	cCol = 3;
}

// destructor
IRT10em::~IRT10em() {
    delete [] I;
}

// sets the default constructor size, used when making arrays of these
static void setnitem(int n) {
  	IRT10em::size = n;
}

// resize IRT10em
void IRT10em::ChangeNitems(int nn) {
    delete [] I;
    int i;

	nitem = nn;
	for (i = 0; i < 64; i++)
		DatFile[i] = kZeroByte;
	I = new ItemStuff [nitem];
}

// resize the output parts of IRT10em that need to know maxscore to size
void IRT10em::ChangeMaxScore(int m) {
	maxscore = m;
	eap.resize(maxscore);
	sd.resize(maxscore);
	wt.resize(maxscore);
	expd.resize(maxscore);
	lk.resize(maxscore);
}  

// graded: input is the item index number i, telling where in IRT10em this goes;
// the slope a and the DblList b containing the b parameters
// (the number of categories is assumed to be in test.I[i].ncat)
// as output, it places in test.I[i].T[x] a DblList of each
// trace line's ordinate at each of the values of IRT10em::theta
void IRT10em::graded (int i) {
	int j;	
// heavily modified version to mimic multilog more closely
// (w.r.t. protection from zero probabilities or differences)
	int q;
	double zl,zr;
	for (q = 0; q < nqpt; q++) {
		zr = I[i].a*(IRT10em::theta[q] - I[i].bj[0]);
		if (zr < -30.0) zr = -30.0;
		if (zr > 30.0) zr = 30.0;
		I[i].T[0][q] = 1 - (1 / (1 + exp(-zr)));
		if (I[i].T[0][q] < 1.0e-20) I[i].T[0][q] = 1.0e-20;
		for (j = 1; j < I[i].ncat - 1; j++) {
			zl = I[i].a*(IRT10em::theta[q] - I[i].bj[j-1]);
			if (zl < -30.0) zl = -30.0;
			if (zl > 30.0) zl = 30.0;
			zr = I[i].a*(IRT10em::theta[q] - I[i].bj[j]);
			if (zr < -30.0) zr = -30.0;
			if (zr > 30.0) zr = 30.0;
			I[i].T[j][q] = (1 / (1 + exp(-zl)))
			 		- (1 / (1 + exp(-zr)));
			if (I[i].T[j][q] < 1.0e-20) I[i].T[j][q] = 1.0e-20;
		}
		zl = I[i].a*(IRT10em::theta[q] - I[i].bj[I[i].ncat - 2]);
		if (zl < -30.0) zl = -30.0;
		if (zl > 30.0) zl = 30.0;
		I[i].T[I[i].ncat - 1][q] =
			 	1 / (1 + exp(-zl));
		if (I[i].T[I[i].ncat - 1][q] < 1.0e-20) I[i].T[I[i].ncat - 1][q] = 1.0e-20;
	}
} // end of graded

// threePL: input is the item index number i, telling where in IRT10em this goes;
// (as currently written, this uses the 1.7)
// trace line's ordinate at each of the values of IRT10em::theta
void IRT10em::threePL (int i) {
// modified to mimic multilog more closely
// (this is about the 'if abs(z) > 20' bits)
	int q;
	double z;
	for (q = 0; q < nqpt; q++) {
		z = I[i].a*1.7*(IRT10em::theta[q] - I[i].b);
		if (z < -20.0) z = -20.0;
		if (z > 20.0) z = 20.0;
		I[i].T[1][q] = I[i].c + ((1 - I[i].c) 
						/ (1 + exp(-z)));
	}
	I[i].T[0] = 1.0 - I[i].T[1];
} // end of threePL

// int_stats: close copy of the LISP function of the same name---
// input is a DblList representing posterior likelihood,
// the IRT10em::theta DblList,
// and a summed score that indicates where in the output the results go---
// computes EAP, S.D., and P for the posterior, as well as (optionally)
// rounded integer scale scores if mu and sigma for those are not 0 and 1
void IRT10em::int_stats(valarray<double>& posterior, int score) {
	
	expd[score] = posterior.sum();
	eap[score] = (posterior * IRT10em::theta).sum() / expd[score];
	sd[score] = sqrt((posterior * 
		(IRT10em::theta - eap[score]) * (IRT10em::theta - eap[score])).sum()/expd[score]);
	wt[score] = 1.0/(sd[score]*sd[score]);
} // end of int_stats

// int_statsR--this is a simpler version that just computes the statistics
// and returns them, for use with Demo()
// ?? 11/24 not sure this is useful any more, but gave it non-conflicting names
void IRT10em::int_statsR(valarray<double>& posterior, 
				double& expdvalue, double& eapvalue, double& sdvalue) {
	
	expdvalue = posterior.sum();
	eapvalue = (posterior * IRT10em::theta).sum() / expdvalue;
	sdvalue = sqrt((posterior * 
		(IRT10em::theta - eapvalue) * (IRT10em::theta - eapvalue)).sum()/expdvalue);
} // end of int_stats


// the first 24 lines comprise a close clone of the LISP we did (not)
// publish in the '95 APM article; it's in the Lab text report
void IRT10em::recursive () {
	int i, ii, k, kk, j;
	vector< valarray<double> > zero_list;
	vector< valarray<double> > lk_temp;
	zero_list.resize(maxscore);
	lk_temp.resize(maxscore);
	lk.resize(maxscore);
	for (int s=0; s<maxscore; s++) {
		zero_list[s].resize(IRT10em::nqpt);
		zero_list[s] = 0.0;
		lk_temp[s].resize(IRT10em::nqpt);
		lk[s].resize(IRT10em::nqpt);
	}
	for (i = 0; i < I[0].ncat; i++) lk[i] = I[0].T[i];
	int I_star = I[0].ncat;
	
	// Begin recursive algorithm
	for (ii = 1; ii < nitem; ii++) {
		for (kk = 0; kk < (I_star + I[ii].ncat - 1); kk++) 
			lk_temp[kk] = zero_list[kk];   
		for (k = 0; k < I[ii].ncat; k++) {
			for (j = 0; j < I_star; j++)
				lk_temp[j + k] = (lk[j] * I[ii].T[k]) + lk_temp[j + k];
		}
		I_star = I_star + (I[ii].ncat - 1);
		for (j = 0; j < I_star; j++) lk[j] = lk_temp[j]; 
	} // the recursive algorithm ends here
	
	// below is stuff added to pick up eaps, weights, etc.
	valarray<double> posterior;
	posterior.resize(IRT10em::nqpt);
	for (i = 0; i < maxscore; i++) {
		posterior = lk[i] * IRT10em::popdist;
		int_stats(posterior, i);
	} // end of loop over scores
} // end of recursive

// Writeit uses cout to write what we hope to place in the output window
void IRT10em::WriteIt () {
	extern ofstream outfile1;
	int i,j;
	
	outfile1 << endl << "Parameter estimates: " << endl;

// print the item parameters
	outfile1 << "Item Type Ncat Parameters:" << endl;
	outfile1 << setprecision(2) << setiosflags(ios::right)
		<< setiosflags(ios::showpoint | ios::fixed);
	for (i = 0; i < nitem; i++) {
		outfile1 << setw(4) << (i + 1) << setw(4) << I[i].kind 
			<< setw(5) << I[i].ncat;
		if (I[i].kind == 1) {
    		outfile1 << setw(6) << I[i].a;
    		for (j = 0; j < (I[i].ncat - 1); j++)
     			outfile1 << setw(6) << I[i].bj[j];
    		outfile1 << endl;
     	}
		if (I[i].kind == 2) 
     		outfile1 << setw(6) << I[i].a
     	 		<< setw(6) << I[i].b << setw(6) << I[i].c << endl;
	} // end of loop over items

} // end of WriteIt
